@@ -25,19 +25,23 @@
 
 from contextlib import contextmanager
 from io import StringIO
+import itertools
 import sys
 import argparse
 import markdown
 import os
+import pathlib
 import subprocess
 from datetime import date, datetime
+from typing import Union
+from textwrap import dedent
 
 import ford.fortran_project
 import ford.sourceform
 import ford.output
-from ford.mdx_math import MathExtension
 import ford.utils
 import ford.pagetree
+from ford.md_environ import EnvironExtension
 
 try:
     from importlib.metadata import version, PackageNotFoundError
@@ -118,19 +122,123 @@ LICENSES = {
 }
 
 
+DEFAULT_SETTINGS = {
+    "alias": [],
+    "author": None,
+    "author_description": None,
+    "author_pic": None,
+    "bitbucket": None,
+    "coloured_edges": False,
+    "copy_subdir": [],
+    "creation_date": "%Y-%m-%dT%H:%M:%S.%f%z",
+    "css": None,
+    "dbg": True,
+    "display": ["public", "protected"],
+    "doc_license": "",
+    "docmark": "!",
+    "docmark_alt": "*",
+    "email": None,
+    "encoding": "utf-8",
+    "exclude": [],
+    "exclude_dir": [],
+    "extensions": ["f90", "f95", "f03", "f08", "f15"],
+    "external": [],
+    "externalize": False,
+    "extra_filetypes": [],
+    "extra_mods": [],
+    "extra_vartypes": [],
+    "facebook": None,
+    "favicon": "default-icon",
+    "fixed_extensions": ["f", "for", "F", "FOR"],
+    "fixed_length_limit": True,
+    "force": False,
+    "fpp_extensions": ["F90", "F95", "F03", "F08", "F15", "F", "FOR"],
+    "github": None,
+    "gitlab": None,
+    "gitter_sidecar": None,
+    "google_plus": None,
+    "graph": False,
+    "graph_dir": None,
+    "graph_maxdepth": "10000",
+    "graph_maxnodes": "1000000000",
+    "hide_undoc": False,
+    "incl_src": True,
+    "include": [],
+    "license": "",
+    "linkedin": None,
+    "lower": False,
+    "macro": [],
+    "mathjax_config": None,
+    "max_frontpage_items": 10,
+    "media_dir": None,
+    "output_dir": "./doc",
+    "page_dir": None,
+    "parallel": 1,
+    "predocmark": ">",
+    "predocmark_alt": "|",
+    "preprocess": True,
+    "preprocessor": "cpp -traditional-cpp -E -D__GFORTRAN__",
+    "print_creation_date": False,
+    "privacy_policy_url": None,
+    "proc_internals": False,
+    "project": "Fortran Program",
+    "project_bitbucket": None,
+    "project_download": None,
+    "project_github": None,
+    "project_gitlab": None,
+    "project_sourceforge": None,
+    "project_url": "",
+    "project_website": None,
+    "quiet": False,
+    "revision": None,
+    "search": True,
+    "sort": "src",
+    "source": False,
+    "src_dir": ["./src"],
+    "summary": None,
+    "terms_of_service_url": None,
+    "twitter": None,
+    "version": None,
+    "warn": False,
+    "website": None,
+    "year": date.today().year,
+}
+
+
+def convert_to_bool(name, option):
+    """Convert value 'option' to a bool, with a nice error message on
+    failure. Expects a list from the markdown meta-data extension"""
+    if len(option) > 1:
+        raise ValueError(
+            f"Could not convert option '{name}' to bool: expected a single value but got a list ({option})"
+        )
+    try:
+        return ford.utils.str_to_bool(option[0])
+    except ValueError:
+        raise ValueError(
+            f"Could not convert option '{name}' to bool: expected 'true'/'false', got: {option[0]}"
+        )
+
+
 def initialize():
     """
     Method to parse and check configurations of FORD, get the project's
     global documentation, and create the Markdown reader.
     """
-    try:
-        import multiprocessing
+    args = get_command_line_arguments()
 
-        ncpus = "{0}".format(multiprocessing.cpu_count())
-    except (ImportError, NotImplementedError):
-        ncpus = "0"
+    # Read in the project-file. This will contain global documentation (which
+    # will appear on the homepage) as well as any information about the project
+    # and settings for generating the documentation.
+    proj_docs = args.project_file.read()
+    directory = os.path.dirname(args.project_file.name)
+
+    return parse_arguments(vars(args), proj_docs, directory)
+
+
+def get_command_line_arguments() -> argparse.Namespace:
+    """Read the command line arguments"""
 
-    # Setup the command-line options and parse them.
     parser = argparse.ArgumentParser(
         "ford",
         description="Document a program or library written in modern Fortran. Any command-line options over-ride those specified in the project file.",
@@ -188,6 +296,7 @@ def initialize():
         "--warn",
         dest="warn",
         action="store_true",
+        default=None,
         help="display warnings for undocumented items",
     )
     parser.add_argument(
@@ -195,12 +304,14 @@ def initialize():
         "--force",
         dest="force",
         action="store_true",
+        default=None,
         help="continue to read file if fatal errors",
     )
     parser.add_argument(
         "--no-search",
         dest="search",
         action="store_false",
+        default=None,
         help="don't process documentation to produce a search feature",
     )
     parser.add_argument(
@@ -208,6 +319,7 @@ def initialize():
         "--quiet",
         dest="quiet",
         action="store_true",
+        default=None,
         help="do not print any description of progress",
     )
     parser.add_argument(
@@ -220,6 +332,7 @@ def initialize():
         "--debug",
         dest="dbg",
         action="store_true",
+        default=None,
         help="display traceback if fatal exception occurs and print faulty line",
     )
     parser.add_argument(
@@ -239,37 +352,54 @@ def initialize():
         "--external_links",
         dest="external",
         action="append",
-        help=""" External projects to link to.
-If an entity is not found in the sources, FORD will try to look it up in
-those external projects. If those have documentation generated by FORD with
-the externalize option, a link will be placed into the documentation wherever
-this entity is referred to.
-FORD will look in the provided paths for a modules.json file.
-""",
+        help="""External projects to link to.
+        If an entity is not found in the sources, FORD will try to look it up in
+        those external projects. If those have documentation generated by FORD with
+        the externalize option, a link will be placed into the documentation wherever
+        this entity is referred to.
+        FORD will look in the provided paths for a modules.json file.
+        """,
     )
-    # Get options from command-line
-    args = parser.parse_args()
+
+    return parser.parse_args()
+
+
+def parse_arguments(
+    command_line_args: dict,
+    proj_docs: str,
+    directory: Union[os.PathLike, str] = os.getcwd(),
+):
+    """Consolidates arguments from the command line and from the project
+    file, and then normalises them how the rest of the code expects
+    """
+
+    try:
+        import multiprocessing
+
+        ncpus = "{0}".format(multiprocessing.cpu_count())
+    except (ImportError, NotImplementedError):
+        ncpus = "0"
+
+    DEFAULT_SETTINGS["parallel"] = ncpus
+
     # Set up Markdown reader
     md_ext = [
         "markdown.extensions.meta",
         "markdown.extensions.codehilite",
         "markdown.extensions.extra",
-        MathExtension(),
-        "md_environ.environ",
+        "mdx_math",
+        EnvironExtension(),
     ]
     md = markdown.Markdown(
         extensions=md_ext, output_format="html5", extension_configs={}
     )
-    # Read in the project-file. This will contain global documentation (which
-    # will appear on the homepage) as well as any information about the project
-    # and settings for generating the documentation.
-    proj_docs = args.project_file.read()
+
     md.convert(proj_docs)
     # Remake the Markdown object with settings parsed from the project_file
     if "md_base_dir" in md.Meta:
         md_base = md.Meta["md_base_dir"][0]
     else:
-        md_base = os.path.dirname(args.project_file.name)
+        md_base = directory
     md_ext.append("markdown_include.include")
     if "md_extensions" in md.Meta:
         md_ext.extend(md.Meta["md_extensions"])
@@ -284,193 +414,27 @@ FORD will look in the provided paths for a modules.json file.
     proj_data = md.Meta
 
     # Get the default options, and any over-rides, straightened out
-    options = [
-        "src_dir",
-        "extensions",
-        "fpp_extensions",
-        "fixed_extensions",
-        "output_dir",
-        "css",
-        "exclude",
-        "external",
-        "externalize",
-        "project",
-        "author",
-        "author_description",
-        "author_pic",
-        "summary",
-        "github",
-        "gitlab",
-        "bitbucket",
-        "facebook",
-        "twitter",
-        "google_plus",
-        "linkedin",
-        "email",
-        "website",
-        "project_github",
-        "project_gitlab",
-        "project_bitbucket",
-        "project_website",
-        "project_download",
-        "project_sourceforge",
-        "project_url",
-        "doc_license",
-        "display",
-        "hide_undoc",
-        "version",
-        "year",
-        "docmark",
-        "predocmark",
-        "docmark_alt",
-        "predocmark_alt",
-        "media_dir",
-        "favicon",
-        "warn",
-        "extra_vartypes",
-        "page_dir",
-        "privacy_policy_url",
-        "terms_of_service_url",
-        "incl_src",
-        "force",
-        "copy_subdir",
-        "alias",
-        "source",
-        "exclude_dir",
-        "macro",
-        "include",
-        "preprocess",
-        "quiet",
-        "search",
-        "lower",
-        "sort",
-        "extra_mods",
-        "dbg",
-        "graph",
-        "graph_maxdepth",
-        "graph_maxnodes",
-        "license",
-        "extra_filetypes",
-        "preprocessor",
-        "creation_date",
-        "print_creation_date",
-        "proc_internals",
-        "coloured_edges",
-        "graph_dir",
-        "gitter_sidecar",
-        "mathjax_config",
-        "parallel",
-        "revision",
-        "fixed_length_limit",
-        "max_frontpage_items",
-        "encoding",
-    ]
-    defaults = {
-        "src_dir": ["./src"],
-        "extensions": ["f90", "f95", "f03", "f08", "f15"],
-        "fpp_extensions": ["F90", "F95", "F03", "F08", "F15", "F", "FOR"],
-        "fixed_extensions": ["f", "for", "F", "FOR"],
-        "output_dir": "./doc",
-        "project": "Fortran Program",
-        "project_url": "",
-        "display": ["public", "protected"],
-        "hide_undoc": "false",
-        "year": date.today().year,
-        "exclude": [],
-        "exclude_dir": [],
-        "copy_subdir": [],
-        "external": [],
-        "externalize": "false",
-        "docmark": "!",
-        "docmark_alt": "*",
-        "predocmark": ">",
-        "predocmark_alt": "|",
-        "alias": [],
-        "favicon": "default-icon",
-        "extra_vartypes": [],
-        "incl_src": "true",
-        "source": "false",
-        "macro": [],
-        "include": [],
-        "preprocess": "true",
-        "preprocessor": "",
-        "proc_internals": "false",
-        "warn": "false",
-        "force": "false",
-        "quiet": "false",
-        "search": "true",
-        "lower": "false",
-        "sort": "src",
-        "extra_mods": [],
-        "dbg": True,
-        "graph": "false",
-        "graph_maxdepth": "10000",
-        "graph_maxnodes": "1000000000",
-        "license": "",
-        "doc_license": "",
-        "extra_filetypes": [],
-        "creation_date": "%Y-%m-%dT%H:%M:%S.%f%z",
-        "print_creation_date": "false",
-        "coloured_edges": "false",
-        "parallel": ncpus,
-        "fixed_length_limit": "true",
-        "max_frontpage_items": 10,
-        "encoding": "utf-8",
-    }
-    listopts = [
-        "extensions",
-        "fpp_extensions",
-        "fixed_extensions",
-        "display",
-        "extra_vartypes",
-        "src_dir",
-        "exclude",
-        "exclude_dir",
-        "external",
-        "macro",
-        "include",
-        "extra_mods",
-        "extra_filetypes",
-        "copy_subdir",
-        "alias",
-    ]
-    # Evaluate paths relative to project file location
-    if args.warn:
-        args.warn = "true"
-    else:
-        del args.warn
-    if args.force:
-        args.force = "true"
-    else:
-        del args.force
-    if args.quiet:
-        args.quiet = "true"
-    else:
-        del args.quiet
-    if not args.search:
-        args.search = "false"
-    else:
-        del args.search
-    for option in options:
-        if hasattr(args, option) and getattr(args, option):
-            proj_data[option] = getattr(args, option)
+    for option, default in DEFAULT_SETTINGS.items():
+        args_option = command_line_args.get(option, None)
+        if args_option is not None:
+            proj_data[option] = args_option
         elif option in proj_data:
             # Think if there is a safe  way to evaluate any expressions found in this list
-            # proj_data[option] = proj_data[option]
-            if option not in listopts:
+            default_type = DEFAULT_SETTINGS.get(option, None)
+            if isinstance(default_type, bool):
+                proj_data[option] = convert_to_bool(option, proj_data[option])
+            elif not isinstance(default_type, list):
+                # If it's not supposed to be a list, then it's
+                # probably supposed to be a single big block of text,
+                # like a description
                 proj_data[option] = "\n".join(proj_data[option])
-        elif option in defaults:
-            proj_data[option] = defaults[option]
-    base_dir = os.path.abspath(os.path.dirname(args.project_file.name))
+        else:
+            proj_data[option] = default
+
+    # Evaluate paths relative to project file location
+    base_dir = pathlib.Path(directory).absolute()
     proj_data["base_dir"] = base_dir
-    for var in ["src_dir", "exclude_dir", "include"]:
-        if var in proj_data:
-            proj_data[var] = [
-                os.path.normpath(
-                    os.path.join(base_dir, os.path.expanduser(os.path.expandvars(p)))
-                )
-                for p in proj_data[var]
-            ]
+
     for var in [
         "page_dir",
         "output_dir",
@@ -478,24 +442,26 @@ FORD will look in the provided paths for a modules.json file.
         "media_dir",
         "css",
         "mathjax_config",
+        "src_dir",
+        "exclude",
+        "exclude_dir",
+        "include",
     ]:
-        if var in proj_data:
-            proj_data[var] = os.path.normpath(
-                os.path.join(
-                    base_dir, os.path.expanduser(os.path.expandvars(proj_data[var]))
-                )
-            )
-    if proj_data["favicon"].strip() != defaults["favicon"]:
-        proj_data["favicon"] = os.path.normpath(
-            os.path.join(
-                base_dir, os.path.expanduser(os.path.expandvars(proj_data["favicon"]))
-            )
-        )
+        if proj_data[var] is None:
+            continue
+        if isinstance(proj_data[var], list):
+            proj_data[var] = [
+                ford.utils.normalise_path(base_dir, p) for p in proj_data[var]
+            ]
+        else:
+            proj_data[var] = ford.utils.normalise_path(base_dir, proj_data[var])
+
+    if proj_data["favicon"].strip() != DEFAULT_SETTINGS["favicon"]:
+        proj_data["favicon"] = ford.utils.normalise_path(base_dir, proj_data["favicon"])
+
     proj_data["display"] = [item.lower() for item in proj_data["display"]]
-    proj_data["incl_src"] = proj_data["incl_src"].lower()
     proj_data["creation_date"] = datetime.now().strftime(proj_data["creation_date"])
-    relative = proj_data["project_url"] == ""
-    proj_data["relative"] = relative
+    proj_data["relative"] = proj_data["project_url"] == ""
     proj_data["extensions"] += [
         ext for ext in proj_data["fpp_extensions"] if ext not in proj_data["extensions"]
     ]
@@ -510,45 +476,25 @@ FORD will look in the provided paths for a modules.json file.
         else:
             extdict[sp[0]] = (sp[1], sp[2])  # (comment_char and lexer_str)
     proj_data["extra_filetypes"] = extdict
+
     # Make sure no src_dir is contained within output_dir
-    for projdir in proj_data["src_dir"]:
-        proj_path = ford.utils.split_path(projdir)
-        out_path = ford.utils.split_path(proj_data["output_dir"])
-        for directory in out_path:
-            if len(proj_path) == 0:
-                break
-            if directory == proj_path[0]:
-                proj_path.remove(directory)
-            else:
-                break
-        else:
-            print(
-                "Error: directory containing source-code {} a subdirectory of output directory {}.".format(
-                    projdir, proj_data["output_dir"]
-                )
+    for srcdir in proj_data["src_dir"]:
+        # In Python 3.9+ we can use pathlib.Path.is_relative_to
+        if proj_data["output_dir"] in (srcdir, *srcdir.parents):
+            raise ValueError(
+                f"Source directory {srcdir} is a subdirectory of output directory {proj_data['output_dir']}."
             )
-            sys.exit(1)
+
     # Check that none of the docmarks are the same
-    if proj_data["docmark"] == proj_data["predocmark"] != "":
-        print("Error: docmark and predocmark are the same.")
-        sys.exit(1)
-    if proj_data["docmark"] == proj_data["docmark_alt"] != "":
-        print("Error: docmark and docmark_alt are the same.")
-        sys.exit(1)
-    if proj_data["docmark"] == proj_data["predocmark_alt"] != "":
-        print("Error: docmark and predocmark_alt are the same.")
-        sys.exit(1)
-    if proj_data["docmark_alt"] == proj_data["predocmark"] != "":
-        print("Error: docmark_alt and predocmark are the same.")
-        sys.exit(1)
-    if proj_data["docmark_alt"] == proj_data["predocmark_alt"] != "":
-        print("Error: docmark_alt and predocmark_alt are the same.")
-        sys.exit(1)
-    if proj_data["predocmark"] == proj_data["predocmark_alt"] != "":
-        print("Error: predocmark and predocmark_alt are the same.")
-        sys.exit(1)
+    docmarks = ["docmark", "predocmark", "docmark_alt", "predocmark_alt"]
+    for first, second in itertools.combinations(docmarks, 2):
+        if proj_data[first] == proj_data[second] != "":
+            raise ValueError(
+                f"{first} ('{proj_data[first]}') and {second} ('{proj_data[second]}') are the same"
+            )
+
     # Add gitter sidecar if specified in metadata
-    if "gitter_sidecar" in proj_data:
+    if proj_data["gitter_sidecar"] is not None:
         proj_docs += """
         <script>
             ((window.gitter = {{}}).chat = {{}}).options = {{
@@ -560,27 +506,26 @@ FORD will look in the provided paths for a modules.json file.
             proj_data["gitter_sidecar"].strip()
         )
     # Handle preprocessor:
-    if proj_data["preprocess"].lower() == "true":
-        if proj_data["preprocessor"]:
-            preprocessor = proj_data["preprocessor"].split()
-        else:
-            preprocessor = ["cpp", "-traditional-cpp", "-E", "-D__GFORTRAN__"]
-
+    if proj_data["preprocess"]:
+        proj_data["preprocessor"] = proj_data["preprocessor"].split()
+        command = proj_data["preprocessor"] + [os.devnull]
         # Check whether preprocessor works (reading nothing from stdin)
         try:
-            devnull = open(os.devnull)
-            subprocess.Popen(
-                preprocessor, stdin=devnull, stdout=devnull, stderr=devnull
-            ).communicate()
-        except OSError as ex:
-            print("Warning: Testing preprocessor failed")
-            print("  Preprocessor command: {}".format(preprocessor))
-            print("  Exception: {}".format(ex))
-            print("  -> Preprocessing turned off")
-            proj_data["preprocess"] = "false"
-        else:
-            proj_data["preprocess"] = "true"
-            proj_data["preprocessor"] = preprocessor
+            subprocess.run(command, check=True, capture_output=True, text=True)
+        except (subprocess.CalledProcessError, OSError) as ex:
+            project_file = command_line_args["project_file"].name
+            exit(
+                dedent(
+                    f"""\
+                    Error: Testing preprocessor command (`{" ".join(command)}`) failed with error:
+                        {ex.stderr.strip() if isinstance(ex, subprocess.CalledProcessError) else ex}
+
+                    If you need to preprocess files, please fix the 'preprocessor' option in '{project_file}'.
+                    Otherwise, please set 'preprocess: False' in '{project_file}'"""
+                )
+            )
+    else:
+        proj_data["fpp_extensions"] = []
 
     # Get the correct license for project license or use value as a custom license value.
     try:
@@ -620,18 +565,6 @@ def main(proj_data, proj_docs, md):
         )
         sys.exit(1)
 
-    # Convert the documentation from Markdown to HTML. Make sure to properly
-    # handle LateX and metadata.
-    if proj_data["relative"]:
-        project.markdown(md, "..")
-    else:
-        project.markdown(md, proj_data["project_url"])
-    project.correlate()
-    if proj_data["relative"]:
-        project.make_links("..")
-    else:
-        project.make_links(proj_data["project_url"])
-
     # Define core macros:
     ford.utils.register_macro("url = {0}".format(proj_data["project_url"]))
     ford.utils.register_macro(
@@ -645,15 +578,22 @@ def main(proj_data, proj_docs, md):
     for alias in proj_data["alias"]:
         ford.utils.register_macro(alias)
 
+    # Convert the documentation from Markdown to HTML. Make sure to properly
+    # handle LateX and metadata.
+    base_url = ".." if proj_data["relative"] else proj_data["project_url"]
+    project.markdown(md, base_url)
+    project.correlate()
+    project.make_links(base_url)
+
     # Convert summaries and descriptions to HTML
     if proj_data["relative"]:
         ford.sourceform.set_base_url(".")
-    if "summary" in proj_data:
+    if proj_data["summary"] is not None:
         proj_data["summary"] = md.convert(proj_data["summary"])
         proj_data["summary"] = ford.utils.sub_links(
             ford.utils.sub_macros(ford.utils.sub_notes(proj_data["summary"])), project
         )
-    if "author_description" in proj_data:
+    if proj_data["author_description"] is not None:
         proj_data["author_description"] = md.convert(proj_data["author_description"])
         proj_data["author_description"] = ford.utils.sub_links(
             ford.utils.sub_macros(
@@ -665,7 +605,7 @@ def main(proj_data, proj_docs, md):
         ford.utils.sub_macros(ford.utils.sub_notes(proj_docs)), project
     )
     # Process any pages
-    if "page_dir" in proj_data:
+    if proj_data["page_dir"] is not None:
         page_tree = ford.pagetree.get_page_tree(
             os.path.normpath(proj_data["page_dir"]), proj_data["copy_subdir"], md
         )
@@ -681,7 +621,7 @@ def main(proj_data, proj_docs, md):
     docs = ford.output.Documentation(proj_data, proj_docs_, project, page_tree)
     docs.writeout()
 
-    if proj_data["externalize"].lower() == "true":
+    if proj_data["externalize"]:
         # save FortranModules to a JSON file which then can be used
         # for external modules
         ford.utils.external(project, make=True, path=proj_data["output_dir"])
@@ -692,11 +632,9 @@ def main(proj_data, proj_docs, md):
 
 def run():
     proj_data, proj_docs, md = initialize()
-    if proj_data["quiet"].lower() == "true":
-        f = StringIO()
-        with stdout_redirector(f):
-            main(proj_data, proj_docs, md)
-    else:
+
+    f = StringIO() if proj_data["quiet"] else sys.stdout
+    with stdout_redirector(f):
         main(proj_data, proj_docs, md)
 
 
