@@ -1,16 +1,8 @@
-import copy
 import pytest
 
 import ford
 
 
-@pytest.fixture
-def restore_macros():
-    old_macros = copy.copy(ford.utils._MACRO_DICT)
-    yield
-    ford.utils._MACRO_DICT = copy.copy(old_macros)
-
-
 def test_sub_macro(restore_macros):
     ford.utils.register_macro("a=b")
     result = ford.utils.sub_macros("|a|")
@@ -33,3 +25,18 @@ def test_register_macro_clash(restore_macros):
 
     with pytest.raises(RuntimeError):
         ford.utils.register_macro("a=c")
+
+
+@pytest.mark.parametrize("string", ["true", "True", "TRUE", "tRuE"])
+def test_str_to_bool_true(string):
+    assert ford.utils.str_to_bool(string)
+
+
+@pytest.mark.parametrize("string", ["false", "False", "FALSE", "fAlSe"])
+def test_str_to_bool_false(string):
+    assert not ford.utils.str_to_bool(string)
+
+
+def test_str_to_bool_already_bool():
+    assert ford.utils.str_to_bool(True)
+    assert not ford.utils.str_to_bool(False)
